//
// Polariser
// Ynformatics Ltd 2016
//
// Using Arduino pro mini 5V 16MHz
//

// pin usage
int digitalInPin = 2; // from photo-transistor for synchronising
int analogInPin = 0;  // also from photo-transistor for calibration
int analogOutPin = 3; // to polariser

int clockPin = 9;    // clock on pin 9  for analog switch
int antiClockPin = 10;// (not used)

String command = "";         // a string to hold incoming command
boolean commandComplete = false;  // whether the command is complete

volatile unsigned long interruptTime = 0;
volatile unsigned long lastInterruptTime = 0;
volatile bool interrupt = false;

unsigned long currentTime;
unsigned long previousTime;

volatile bool inputHigh = false;

int frameTime = 60;

volatile int calibrateStepLength = 5;

int minCalibratePulseLength = 50;
int maxCalibratePulseLength = 2000;

int fortyFiveValue = 100;
bool manual = false;
int analogManual = 0;

typedef enum PolariserState { ZERO, FORTYFIVE, NINETY, CALIBRATING};
PolariserState polariserState = ZERO;

void setup() 
{
  Serial.begin(9600);

  // Clock for switcher
  StartOscillator(); 

  // Timer 2: PWM 3 & 11 @ 32 kHz
  TCCR2B = 0x01;   

  // Interrupt on sync pulse
  attachInterrupt(digitalPinToInterrupt(digitalInPin), isr, RISING);
}

void isr()
{ 
    lastInterruptTime = interruptTime;
    interruptTime = millis();
    interrupt = true;   
    inputHigh = true;
}

void loop() 
{
  if (commandComplete) 
  {
     HandleCommand();
  } 
  
  currentTime = millis();
  
  if(manual)
  {
    // for testing
    analogWrite( analogOutPin, analogManual); 
  }
  else
  {
    if(polariserState == CALIBRATING && (currentTime - previousTime >= maxCalibratePulseLength))
    {
       polariserState = NINETY;   
    }
    
    if(polariserState != CALIBRATING && digitalRead(digitalInPin) == LOW)
    {
       inputHigh = false;  
    }

    // pseudo state machine

    // calibration pulse received
    if(polariserState != CALIBRATING && inputHigh && currentTime > interruptTime && currentTime - interruptTime > minCalibratePulseLength)
    {  
      Calibrate();
      polariserState = CALIBRATING;
    }

    // synchronisation pulse received
    if(polariserState != CALIBRATING &&  interrupt)
    {
      interrupt = false;
      polariserState = NINETY;
      previousTime = currentTime;
      analogWrite( analogOutPin, 0);
    }
  
    if(interruptTime > lastInterruptTime && interruptTime - lastInterruptTime  < 400 && interruptTime - lastInterruptTime  > 30)
    {
       frameTime = (interruptTime - lastInterruptTime) / 3; 
    }

    // cycle through polarisation levels
    if((polariserState == NINETY) && (currentTime - previousTime >= frameTime))
    {
      polariserState = FORTYFIVE;  // next state
      previousTime = currentTime;  
      analogWrite( analogOutPin, fortyFiveValue);
    }
    else if ((polariserState == FORTYFIVE) && (currentTime - previousTime >= frameTime))
    {
      polariserState = ZERO;   
      previousTime = currentTime;   
      analogWrite( analogOutPin, 255);
    }  
    else if ((polariserState == ZERO) && (currentTime - previousTime >= frameTime))
    {
      polariserState = NINETY;
      previousTime = currentTime;
      analogWrite( analogOutPin, 0);
    }
  }
}

void Calibrate()
{
  int best = 1023;
  int val;
  
  for(int i = 0; i < 255; i+=2)
  {
    analogWrite( analogOutPin, i);
    delay(calibrateStepLength);
    val = analogRead(analogInPin);
    
    //Serial.print(i); Serial.print(" "); Serial.println(val);
    
    if(i > 10 && i < 240 && val < best) // discard extremes
    {
      best = val;
      fortyFiveValue = i;     
    }
  }
  
  //Serial.print("best "); Serial.print(fortyFiveValue); Serial.print(" "); Serial.println(best);
}

void serialEvent() 
{
  while (Serial.available()) 
  {
    char inChar = (char)Serial.read(); 
   
    if (inChar == '\n') 
    {
      commandComplete = true;     
    } 
    else   
      command += inChar;
  }
}

// simple serial interface for testing
void HandleCommand()
{
     if(command.startsWith("f"))
     {
        frameTime = command.substring(1).toInt();            
     }   
     else if(command.startsWith("c"))
     {
        calibrateStepLength = command.substring(1).toInt();            
     }   
     else if(command.startsWith("m"))
     {
        analogManual = command.substring(1).toInt(); 
        manual = true;         
     }   
     else if(command.startsWith("n"))
     {
        manual = false;         
     }   
     else
     {
        // show help
        Serial.println("f20 - frameTime 20 milisecs");
        Serial.println("c20 - calibrateDelay 20 milisecs");
        Serial.println("m20 - manual output 20 (0 - 255)");
        Serial.println("n - cancel manual");
     }
          
     command = "";
     commandComplete = false;  
}

void StartOscillator()
{
  // Times in microseconds, converted to timer ticks
  //  ticks depend on 16 MHz clock, so ticks = 62.5 ns
   
  #define TIMER1_PRESCALE   1     // clock prescaler value
  #define TCCR1B_CS20       0x01  // CS2:0 bits = prescaler selection
   
  #define PERIOD_US   2000  //2000 == 500Hz 
  #define PERIOD_TICKS (microsecondsToClockCycles(PERIOD_US / 2) / TIMER1_PRESCALE)
  
  // Configure Timer 1 for Freq-Phase Correct PWM
  //   Timer 1 + output on OC1A, chip pin 15, Arduino PWM9
  //   Timer 1 - output on OC1B, chip pin 16, Arduino PWM10
   
   analogWrite(clockPin,128);    // let Arduino setup do its thing
   analogWrite(antiClockPin,128);
   
   TCCR1B = 0x00;                 // stop Timer1 clock for register updates
   
  // Clear OC1A on match, P-F Corr PWM Mode: lower WGM1x = 00
   TCCR1A = 0x80 | 0x00;
   
  // If push-pull drive, set OC1B on match
   TCCR1A |= 0x30;
   
   ICR1 = PERIOD_TICKS;           // PWM period
   OCR1A = PERIOD_TICKS / 2;      // ON duration = drive pulse width = 50% duty cycle
   OCR1B = PERIOD_TICKS / 2;      //  ditto - use separate load due to temp buffer reg
   TCNT1 = OCR1A - 1;             // force immediate OCR1x compare on next tick
   
  // upper WGM1x = 10, Clock Sel = prescaler, start Timer 1 running
   TCCR1B = 0x10 | TCCR1B_CS20;

}
